#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

from oslo_serialization import jsonutils

from tacker.objects import fields
from tacker.tests.functional.sol.vnflcm import base as vnflcm_base
from tacker.tests.functional.sol_multi_tenant.vnflcm import base
from tacker.tests.functional.sol_multi_tenant.vnflcm import fake_vnflcm
import tempfile
import time


class VnfLcmWithMultiTenant(base.BaseVnfLcmMultiTenantTest):

    VNF_PACKAGE_DELETE_TIMEOUT = 120

    @classmethod
    def setUpClass(cls):
        super(VnfLcmWithMultiTenant, cls).setUpClass()

        # ModifyVNF specific image create.
        is_setup_error_tenant1 = cls._modify_vnf_specific_image_create(
            cls.glance_client_tenant1)
        if is_setup_error_tenant1:
            cls.is_setup_error = True
            return

        is_setup_error_tenant2 = cls._modify_vnf_specific_image_create(
            cls.glance_client_tenant2)
        if is_setup_error_tenant2:
            cls.is_setup_error = True
            return

    @classmethod
    def _modify_vnf_specific_image_create(cls, glance_clt):
        is_setup_error = False
        images = cls._list_glance_image()
        if len(images) == 0:
            is_setup_error = True
            return is_setup_error

        for image in images:
            specific_image_name = f'{image.name}{2}'
            image_data = {
                "min_disk": image.min_disk,
                "min_ram": image.min_ram,
                "disk_format": image.disk_format,
                "container_format": image.container_format,
                "visibility": image.visibility,
                "name": specific_image_name}

            try:
                images = cls._list_glance_image(specific_image_name)
                if len(images) == 1:
                    break

                _, body = glance_clt.http_client.get(
                    f'{glance_clt.http_client.get_endpoint()}{image.file}')

                with tempfile.TemporaryFile('w+b') as f:
                    for content in body:
                        f.write(content)
                    cls._create_glance_image(image_data, f.read())
            except Exception as e:
                print("Fail, Modify-VNF specific image create.", e, flush=True)
                is_setup_error = True
                return is_setup_error

            return is_setup_error

    def _wait_show_subscription(self, subscription_id, tacker_client):
        # wait for subscription creation
        timeout = vnflcm_base.VNF_SUBSCRIPTION_TIMEOUT
        start_time = int(time.time())
        while True:
            resp, body = self._show_subscription(subscription_id,
                    tacker_client)
            if resp.ok or resp.status_code == 404:
                return resp, body

            if ((int(time.time()) - start_time) > timeout):
                if resp:
                    resp.raise_for_status()
                raise TimeoutError("Failed to show_subscription")

            time.sleep(1)

    def _delete_vnf_package(self, package_uuid, http_client):
        url = os.path.join(self.base_url, package_uuid)
        resp, _ = http_client.do_request(url, "DELETE")
        self.assertEqual(204, resp.status_code)

    def _wait_for_delete(self, package_uuid, http_client):
        show_url = os.path.join(self.base_url, package_uuid)
        timeout = self.VNF_PACKAGE_DELETE_TIMEOUT
        start_time = int(time.time())
        while True:
            resp, body = http_client.do_request(show_url, "GET")
            if (404 == resp.status_code):
                return resp, body

            if (int(time.time()) - start_time) > timeout:
                raise Exception("Failed to delete package")
            time.sleep(1)

    def _disable_operational_state(self, package_uuid, http_client):
        update_req_body = jsonutils.dumps({
            "operationalState": "DISABLED"})

        resp, _ = http_client.do_request(
            '{base_path}/{id}'.format(id=package_uuid,
                                      base_path=self.base_url),
            "PATCH", content_type='application/json',
            body=update_req_body)
        self.assertEqual(200, resp.status_code)

    def assert_vnf_package_usage_state(
            self,
            vnf_package_info,
            expected_usage_state=fields.PackageUsageStateType.IN_USE):
        self.assertEqual(
            expected_usage_state,
            vnf_package_info['usageState'])

    def assert_create_vnf(self, resp, vnf_instance, vnf_pkg_id,
            tacker_client, fake_server_manager):
        super().assert_create_vnf(resp, vnf_instance,
            fake_server_manager)

        resp, vnf_pkg_info = vnflcm_base._show_vnf_package(
            tacker_client, vnf_pkg_id)
        self.assert_vnf_package_usage_state(vnf_pkg_info)

    def _vnf_instance_wait_until_fail_detected(self, id,
            instantiation_state=fields.VnfInstanceState.NOT_INSTANTIATED,
            timeout=vnflcm_base.VNF_INSTANTIATE_ERROR_WAIT):
        time.sleep(timeout)
        _, body = self._show_vnf_instance(id)
        if body['instantiationState'] != instantiation_state:
            error = ("Vnf instance %(id)s status is %(current)s, "
                     "expected status should be %(expected)s")
            self.fail(error % {"id": id,
                "current": body['instantiationState'],
                "expected": instantiation_state})

    def test_subscription_functionality(self):
        """Test subscription operations with member role users.

        In this test case, we do following steps.
        Note: User A belongs to Tenant 1(t1).
              User B belongs to Tenant 2(t2).
            - Create subscription.
              - User A registers Subscription A(Notification Server A).
              - User B registers Subscription B(Notification Server B).
            - Show Subscription
              - User A only gets information about Subscription A.
              - User B only gets information about Subscription B.
            - List Subscription
              - User A gets subscription list and confirms only
                Subscription A is output.
              - User B gets subscription list and confirms only
                Subscription B is output.
            - Delete Subscription
              - User A deletes Subscription A.
              - User B deletes Subscription B.
        TODO(manpreetk): Only positive test cases are validated in
        Y-release.
        Negative test cases
              - User A fails to delete Subscription B.
              - User B fails to delete Subscription A.
        Validation of negative test cases would require design changes
        in Fake NFVO server, which could be implemented in the upcoming
        cycle.
        """
        # Create subscription
        # User A registers Subscription A.
        callback_url = os.path.join(vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        req_body = fake_vnflcm.Subscription.make_create_request_body(
            'http://localhost:{}{}'.format(
                base.FAKE_SERVER_MANAGER_T1.SERVER_PORT_T1,
                callback_url))
        resp_t1, resp_body_t1 = self._register_subscription(req_body,
            self.http_client_tenant1)
        self.assertEqual(201, resp_t1.status_code)
        self.assert_http_header_location_for_subscription(
            resp_t1.headers)
        self.assert_notification_get(callback_url,
            base.FAKE_SERVER_MANAGER_T1)
        subscription_id_t1 = resp_body_t1.get('id')

        # User B registers Subscription B
        callback_url = os.path.join(vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        req_body_t2 = fake_vnflcm.Subscription.make_create_request_body(
            'http://localhost:{}{}'.format(
                base.FAKE_SERVER_MANAGER_T2.SERVER_PORT_T2,
                callback_url))
        resp_t2, resp_body_t2 = self._register_subscription(
            req_body_t2, self.http_client_tenant2)
        self.assertEqual(201, resp_t2.status_code)
        self.assert_http_header_location_for_subscription(
            resp_t2.headers)
        self.assert_notification_get(callback_url,
            base.FAKE_SERVER_MANAGER_T2)
        subscription_id_t2 = resp_body_t2.get('id')

        # Show Subscription
        # User A gets information for Subscription A
        resp_t1, resp_body_show_t1 = self._wait_show_subscription(
            subscription_id_t1, self.tacker_client_t1)
        self.assert_subscription_show(resp_t1, resp_body_show_t1)

        # User B gets information for Subscription B
        resp_t2, resp_body_show_t2 = self._wait_show_subscription(
            subscription_id_t2, self.tacker_client_t2)
        self.assert_subscription_show(resp_t2, resp_body_show_t2)

        # List Subscription
        # User A gets subscription list
        resp, _ = self._list_subscription(self.tacker_client_t1)
        self.assertEqual(200, resp_t1.status_code)

        # Confirm subscription A
        filter_expr = {
            'filter': "filter=(eq,id,{})".format(
                resp_body_show_t1.get('id'))}
        resp, subscription_body_t1 = self._list_subscription_filter(
            self.http_client_tenant1,
            params=filter_expr)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(1, len(subscription_body_t1))

        # User B gets subscription list
        resp_t2, _ = self._list_subscription(
            self.tacker_client_t2)
        self.assertEqual(200, resp_t2.status_code)

        # Confirm subscription B
        filter_expr = {
            'filter': "filter=(eq,id,{})".format(
                resp_body_show_t2.get('id'))}
        resp, subscription_body_t2 = self._list_subscription_filter(
            self.http_client_tenant2,
            params=filter_expr)
        self.assertEqual(200, resp.status_code)
        self.assertEqual(1, len(subscription_body_t2))

        # Delete subscription
        # User A deletes Subscription A
        self.addCleanup(self._delete_subscription,
            subscription_id_t1, self.tacker_client_t1)

        # User B deletes Subscription B
        self.addCleanup(self._delete_subscription,
            subscription_id_t2, self.tacker_client_t2)

    def test_vnf_package_functionality(self):
        """Test VNF package operations with member role users.

        In this test case, we do following steps.
        Note: User A belongs to Tenant 1.
              User B belongs to Tenant 2.
            - Create and Upload VNF Package
              - User A creates VNF Package A.
              - User A uploads VNF Package A.
              - User B creates VNF Package B.
              - User B uploads VNF Package B.
            - List VNF Package
              - User A gets VNF package list and confirms only
                VNF Package A is output.
              - User B gets VNF package list and confirms only
                VNF Package B is output.
            - Show VNF Package
              - User A only gets information about VNF Package A.
              - User B only gets information about VNF Package B.
            - Delete VNF Package
              - User A deletes VNF Package A.
              - User B deletes VNF Package B.
        TODO(manpreetk): Only positive test cases are validated in
        Y-release.
        Negative test cases
              - User A fails to delete VNF Package B.
              - User B fails to delete VNF Package A.
        Validation of negative test cases would require design changes
        in Fake NFVO server, which could be implemented in the upcoming
        cycle.
        """
        # Create and Upload VNF Package
        # User A creates VNF Package A
        sample_name = 'mt_functional1'
        csar_package_path = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../../etc/samples/etsi/nfv",
                sample_name))
        tempname, _ = vnflcm_base._create_csar_with_unique_vnfd_id(
            csar_package_path)

        # User A uploads VNF Package A
        vnf_pkg_id, vnfd_id = vnflcm_base._create_and_upload_vnf_package(
            self.tacker_client_t1, user_defined_data={
                "key": sample_name}, temp_csar_path=tempname)

        # User B creates VNF Package B
        sample_name_t2 = 'mt_functional1'
        csar_package_path_t2 = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../../etc/samples/etsi/nfv",
                sample_name_t2))
        tempname_t2, _ = vnflcm_base._create_csar_with_unique_vnfd_id(
            csar_package_path_t2)

        # User B uploads VNF Package B
        vnf_pkg_id2, vnfd_id2 = vnflcm_base._create_and_upload_vnf_package(
            self.tacker_client_t2, user_defined_data={
                "key": sample_name_t2}, temp_csar_path=tempname_t2)

        # List VNF Package
        # User A gets VNF package list and confirms only VNF Package A
        # is output.
        resp, body = self.http_client_tenant1.do_request(
            self.base_url, "GET")
        self.assertEqual(200, resp.status_code)

        package_id = [obj['id'] for obj in body]
        self.assertIn(vnf_pkg_id, package_id)

        # User B gets VNF package list and confirms only VNF Package B
        # is output.
        resp_t2, body_t2 = self.http_client_tenant2.do_request(
            self.base_url, "GET")
        self.assertEqual(200, resp_t2.status_code)

        package_id = [obj['id'] for obj in body_t2]
        self.assertIn(vnf_pkg_id2, package_id)

        # Show VNF Package
        # User A only gets information about VNF Package A
        show_url = os.path.join(self.base_url, vnf_pkg_id)
        resp, body = self.http_client_tenant1.do_request(
            show_url, "GET")
        self.assertEqual(200, resp.status_code)

        # User B only gets information about VNF Package B
        show_url_t2 = os.path.join(self.base_url, vnf_pkg_id2)
        resp_t2, body_t2 = self.http_client_tenant2.do_request(
            show_url_t2, "GET")
        self.assertEqual(200, resp_t2.status_code)

        # Delete VNF Package
        # User A deletes VNF Package A
        self._disable_operational_state(vnf_pkg_id,
            self.http_client_tenant1)
        self._delete_vnf_package(vnf_pkg_id, self.http_client_tenant1)
        self._wait_for_delete(vnf_pkg_id, self.http_client_tenant1)

        # User B deletes VNF Package B
        self._disable_operational_state(vnf_pkg_id2,
            self.http_client_tenant2)
        self._delete_vnf_package(vnf_pkg_id2, self.http_client_tenant2)
        self._wait_for_delete(vnf_pkg_id2, self.http_client_tenant2)

    def test_vnf_instantiation_by_vim_of_different_tenant_and_role(self):
        """Test VNF instantiation by VIM of differnt tenant.

        In this test case, we do following steps.
        Note: User A is an admin role user belongs to Tenant 1.
              User B is a non-admin role user belongs to Tenant 2.
            - Create VNF Instance
              - User B creates VNF Instance B using VNF Package B.
            - Instantiate VNF
              - User A fails to instantiates VNF Instance B, both
                VNF and VIM belong to different tenant.
        """
        # Pre-Setting
        # User B registers Subscription B.
        callback_url = os.path.join(vnflcm_base.MOCK_NOTIFY_CALLBACK_URL,
            self._testMethodName)
        req_body_t2 = fake_vnflcm.Subscription.make_create_request_body(
            'http://localhost:{}{}'.format(
                base.FAKE_SERVER_MANAGER_T2.SERVER_PORT_T2,
                callback_url))
        resp_t2, resp_body_t2 = self._register_subscription(
            req_body_t2, self.http_client_tenant2)
        self.assertEqual(201, resp_t2.status_code)
        self.assert_http_header_location_for_subscription(
            resp_t2.headers)
        self.assert_notification_get(callback_url,
            base.FAKE_SERVER_MANAGER_T2)
        subscription_id_t2 = resp_body_t2.get('id')
        self.addCleanup(
            self._delete_subscription,
            subscription_id_t2,
            self.tacker_client_t2)

        # Create and Upload VNF Package
        # User B creates VNF Package B
        sample_name_t2 = 'mt_functional1'
        csar_package_path_t2 = os.path.abspath(
            os.path.join(
                os.path.dirname(__file__),
                "../../../etc/samples/etsi/nfv",
                sample_name_t2))
        tempname_t2, _ = vnflcm_base._create_csar_with_unique_vnfd_id(
            csar_package_path_t2)

        # User B uploads VNF Package B
        vnf_pkg_id, vnfd_id = vnflcm_base._create_and_upload_vnf_package(
            self.tacker_client_t2, user_defined_data={
                "key": sample_name_t2}, temp_csar_path=tempname_t2)

        # Post Setting: Reserve deleting VNF package.
        self.addCleanup(vnflcm_base._delete_vnf_package,
            self.tacker_client_t2, vnf_pkg_id)

        # Create VNF Instance
        # User B creates VNF Instance B using VNF Package B
        resp_t2, vnf_instance_t2 = self._create_vnf_instance_from_body(
            fake_vnflcm.VnfInstances.make_create_request_body(vnfd_id),
            self.http_client_tenant2)
        vnf_instance_id_t2 = vnf_instance_t2.get('id')
        self._wait_lcm_done(vnf_instance_id=vnf_instance_id_t2,
            fake_server_manager=base.FAKE_SERVER_MANAGER_T2)
        self.assert_create_vnf(resp_t2, vnf_instance_t2,
            vnf_pkg_id,
            self.tacker_client_t2,
            base.FAKE_SERVER_MANAGER_T2)

        # Instantiate VNF instance
        # User A fails to instantiate VNF Instance B as both VIM and VNF
        # belongs to differernt tenants
        request_body = fake_vnflcm.VnfInstances.make_inst_request_body(
            'nfv_user',
            self.vim['tenant_id'], self.ext_networks,
            self.ext_mngd_networks,
            self.ext_link_ports, self.ext_subnets)
        resp, _ = self._instantiate_vnf_instance(vnf_instance_id_t2,
            request_body, self.http_client)
        self.assertEqual(202, resp.status_code)
        self._vnf_instance_wait_until_fail_detected(vnf_instance_id_t2)



# Copyright 2016 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""set-mandatory-columns-not-null

Revision ID: 354de64ba129
Revises: b07673bb8654
Create Date: 2016-06-02 10:05:22.299780

"""

# flake8: noqa: E402

# revision identifiers, used by Alembic.
revision = '354de64ba129'
down_revision = 'b07673bb8654'

from alembic import op
import sqlalchemy as sa


def upgrade(active_plugins=None, options=None):
    for table in ['devices', 'devicetemplates', 'vims', 'servicetypes']:
        op.alter_column(table,
                        'tenant_id',
                        existing_type=sa.String(64),
                        nullable=False)



# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# TODO(yamahata): once unittests are impletemted, move this there
from oslo_log import log as logging
from oslo_utils import uuidutils

from tacker.common import log
from tacker.vnfm.infra_drivers import abstract_driver


LOG = logging.getLogger(__name__)


class VnfNoop(abstract_driver.VnfAbstractDriver):

    """Noop driver of hosting vnf for tests."""

    def __init__(self):
        super(VnfNoop, self).__init__()
        self._instances = set()

    def get_type(self):
        return 'noop'

    def get_name(self):
        return 'noop'

    def get_description(self):
        return 'Tacker infra noop driver'

    @log.log
    def create(self, **kwargs):
        instance_id = uuidutils.generate_uuid()
        self._instances.add(instance_id)
        return instance_id

    @log.log
    def create_wait(self, plugin, context, vnf_dict, vnf_id):
        pass

    @log.log
    def update(self, plugin, context, vnf_id, vnf_dict, vnf):
        if vnf_id not in self._instances:
            LOG.debug('not found')
            raise ValueError('No instance %s' % vnf_id)

    @log.log
    def update_wait(self, plugin, context, vnf_id):
        pass

    @log.log
    def delete(self, plugin, context, vnf_id):
        self._instances.remove(vnf_id)

    @log.log
    def delete_wait(self, plugin, context, vnf_id):
        pass

    def get_resource_info(self, plugin, context, vnf_info, auth_attr,
                          region_name=None):
        return {'noop': {'id': uuidutils.generate_uuid(), 'type': 'noop'}}

    def heal_vdu(self, plugin, context, vnf_dict, heal_request_data):
        pass

    def pre_instantiation_vnf(self, context, vnf_instance,
                              vim_connection_info, image_data):
        pass

    def delete_vnf_instance_resource(self, context, vnf_instance,
            vim_connection_info, vnf_resource):
        pass

    def instantiate_vnf(self, context, vnf_instance, vnfd_dict,
                        vim_connection_info, instantiate_vnf_req,
                        grant_response):
        pass

    def post_vnf_instantiation(self, context, vnf_instance,
                               vim_connection_info, instantiate_vnf_req):
        pass

    def heal_vnf(self, context, vnf_instance, vim_connection_info,
                 heal_vnf_request):
        pass

    def heal_vnf_wait(self, context, vnf_instance, vim_connection_info):
        pass

    def post_heal_vnf(self, context, vnf_instance, vim_connection_info,
                      heal_vnf_request):
        pass

    def change_ext_conn_vnf(self, context, vnf_instance, vnf_dict,
                            vim_connection_info, change_ext_conn_req):
        pass

    def change_ext_conn_vnf_wait(self, context, vnf_instance,
                                 vim_connection_info):
        pass

    def post_change_ext_conn_vnf(self, context, vnf_instance,
                                 vim_connection_info):
        pass



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# flake8: noqa: E402

"""add_vnf_pkg_id_to_vnf_instances

Revision ID: f9bc96967462
Revises: e06fbdc90a32
Create Date: 2020-09-11 22:41:25.799693

"""

# revision identifiers, used by Alembic.
revision = 'f9bc96967462'
down_revision = 'e06fbdc90a32'

from alembic import op
import sqlalchemy as sa


from tacker.db import migration
from tacker.db import types


def upgrade(active_plugins=None, options=None):
    op.add_column('vnf_instances', sa.Column('vnf_pkg_id',
        types.Uuid(length=36), nullable=False))



# Copyright (c) 2012 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from unittest import mock

from oslo_config import cfg

from tacker.common import config
from tacker.tests import base


class ConfigurationTest(base.BaseTestCase):

    def test_defaults(self):
        self.assertEqual('0.0.0.0', cfg.CONF.bind_host)
        self.assertEqual(9890, cfg.CONF.bind_port)
        self.assertEqual('api-paste.ini.test', cfg.CONF.api_paste_config)
        self.assertEqual('unit/extensions', cfg.CONF.api_extensions_path)
        self.assertEqual('keystone', cfg.CONF.auth_strategy)
        self.assertTrue(cfg.CONF.allow_bulk)
        self.assertFalse(cfg.CONF.allow_pagination)
        self.assertFalse(cfg.CONF.allow_sorting)
        self.assertEqual('-1', cfg.CONF.pagination_max_limit)
        relative_dir = os.path.join(os.path.dirname(__file__),
                                    '..', '..', '..', '..')
        absolute_dir = os.path.abspath(relative_dir)
        self.assertEqual(absolute_dir, cfg.CONF.state_path)
        self.assertEqual('tacker', cfg.CONF.control_exchange)
        self.assertEqual('sqlite://', cfg.CONF.database.connection)

    def test_load_paste_app_not_found(self):
        self.config(api_paste_config='no_such_file.conf')
        with mock.patch.object(cfg.CONF, 'find_file', return_value=None) as ff:
            e = self.assertRaises(cfg.ConfigFilesNotFoundError,
                                  config.load_paste_app, 'app')
            ff.assert_called_once_with('no_such_file.conf')
            self.assertEqual(['no_such_file.conf'], e.config_files)

    @mock.patch('paste.deploy.loadapp')
    def test_load_paste_app(self, mock_deploy):
        mock_deploy.return_value = 'test'
        self.config(api_paste_config='api-paste.ini.test')
        config.load_paste_app('tacker')
        config_path = os.path.abspath(cfg.CONF.find_file(
            cfg.CONF.api_paste_config))
        mock_deploy.assert_called_once_with('config:%s' % config_path,
                                            name='tacker')

    def test_load_paste_app_runtime_error(self):
        self.config(api_paste_config='api-paste.ini.test')
        from paste import deploy
        with mock.patch.object(deploy, 'loadapp', side_effect=RuntimeError):
            self.assertRaises(RuntimeError, config.load_paste_app, 'tacker')



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import re

from flask import Flask
from flask import request
from keystoneauth1.identity import v3
from keystoneauth1 import session
from novaclient import client as nova_client
from tackerclient.v1_0 import client as tacker_client
import yaml

from tacker.tests import constants
from tacker.tests.functional.sol_separated_nfvo.vnflcm.fake_grant import Grant
from tacker.tests.utils import read_file


class GrantServer:
    app = Flask(__name__)

    def __init__(self):
        self.client = self.tackerclient()
        self.nova_client = self.novaclient()

    @staticmethod
    def log_http_request():
        message = "Method:{0}, Url:{1}, Headers:{2}, Body:{3}"
        body = ""
        ct = "{0}".format(request.headers.get("Content-type"))
        print(ct)
        if len(request.get_data()) > 0 and not re.match(
                ".*?application/zip.*?", ct):
            body = request.get_data().decode("utf-8")
        hs = ""
        ff = "{0}:{1};"
        for k, v in request.headers.items():
            hs += ff.format(k, v)
        message = message.format(request.method, request.url, hs, body)
        print(message)

    @staticmethod
    def log_http_response(resp):
        message = "Status:{0}, Body:{1}"
        body = ""

        if len(resp.get_data()) > 0:
            try:
                body = resp.get_data().decode("utf-8")
            except AttributeError:
                body = "binary file."
        message = message.format(resp.status_code, body)
        print(message)
        return resp

    def get_auth_session(self):
        vim_params = self.get_credentials()
        auth = v3.Password(
            auth_url=vim_params['auth_url'],
            username=vim_params['username'],
            password=vim_params['password'],
            project_name=vim_params['project_name'],
            user_domain_name=vim_params['user_domain_name'],
            project_domain_name=vim_params['project_domain_name'])
        verify = 'True' == vim_params.pop('cert_verify', 'False')
        auth_ses = session.Session(auth=auth, verify=verify)
        return auth_ses

    def get_credentials(self):
        vim_params = yaml.safe_load(read_file('local-vim.yaml'))
        vim_params['auth_url'] += '/v3'
        return vim_params

    def tackerclient(self):
        auth_session = self.get_auth_session()
        return tacker_client.Client(session=auth_session, retries=5)

    def novaclient(self):
        vim_params = self.get_credentials()
        auth = v3.Password(auth_url=vim_params['auth_url'],
            username=vim_params['username'],
            password=vim_params['password'],
            project_name=vim_params['project_name'],
            user_domain_name=vim_params['user_domain_name'],
            project_domain_name=vim_params['project_domain_name'])
        verify = 'True' == vim_params.pop('cert_verify', 'False')
        auth_ses = session.Session(auth=auth, verify=verify)
        return nova_client.Client(constants.NOVA_CLIENT_VERSION,
                                  session=auth_ses)

    def list_zone(self):
        try:
            zone = self.nova_client.services.list()
        except nova_client.exceptions.ClientException:
            print("availability zone does not exists.", flush=True)
            return []
        return zone

    def get_vim(self):
        vim_list = self.client.list_vims()
        vim = self.get_vim_specified(vim_list, 'openstack-admin-vim')
        if not vim:
            assert False, "vim_list is Empty: Default VIM is missing"
        return vim

    def get_vim_specified(self, vim_list, vim_name):
        if len(vim_list.values()) == 0:
            assert False, "vim_list is Empty: Default VIM is missing"

        for vim_list in vim_list.values():
            for vim in vim_list:
                if vim['name'] == vim_name:
                    return vim
        return None


@GrantServer.app.route('/grant/v1/grants', methods=['POST'])
def grant():
    body = request.json
    request_body = Grant.convert_body_to_dict(body)
    glance_image = Grant.get_sw_image("functional5")
    flavour_vdu_dict = Grant.get_compute_flavor("functional5")
    availability_zone_info = GrantServer().list_zone()
    zone_name_list = list(set(
        [zone.zone for zone in availability_zone_info
         if zone.binary == 'nova-compute']))
    vim = GrantServer().get_vim()
    if request_body['operation'] == 'INSTANTIATE':
        return Grant.make_inst_response_body(
            body, vim['tenant_id'], glance_image, flavour_vdu_dict,
            zone_name_list)
    if request_body['operation'] == 'SCALE':
        return Grant.make_scale_response_body(
            body, vim['tenant_id'], glance_image, flavour_vdu_dict,
            zone_name_list)
    if request_body['operation'] == 'HEAL':
        return Grant.make_heal_response_body(
            body, vim['tenant_id'], glance_image, flavour_vdu_dict,
            zone_name_list)
    if request_body['operation'] == 'CHANGE_EXT_CONN':
        return Grant.make_change_ext_conn_response_body(
            body, vim['tenant_id'], zone_name_list)
    if request_body['operation'] == 'TERMINATE':
        return Grant.make_term_response_body(body)


# Start Fake_Grant_Server for manual test
GrantServer.app.before_request(GrantServer.log_http_request)
GrantServer.app.after_request(GrantServer.log_http_response)
GrantServer.app.run(host="127.0.0.1", port=9990, debug=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from toscaparser import tosca_template
from toscaparser.utils import yamlparser
from translator.hot import tosca_translator
import yaml

from tacker._i18n import _
from tacker.common import exceptions
from tacker.common import log
from tacker.extensions import common_services as cs
from tacker.extensions import vnfm
from tacker.plugins.common import constants
from tacker.tosca import utils as toscautils


LOG = logging.getLogger(__name__)
CONF = cfg.CONF

OPTS = [
    cfg.DictOpt('flavor_extra_specs',
               default={},
               help=_("Flavor Extra Specs")),
]

CONF.register_opts(OPTS, group='openstack_vim')

HEAT_VERSION_INCOMPATIBILITY_MAP = {'OS::Neutron::Port': {
    'port_security_enabled': 'value_specs', }, }

HEAT_TEMPLATE_BASE = """
heat_template_version: 2013-05-23
"""

ALARMING_POLICY = 'tosca.policies.tacker.Alarming'
SCALING_POLICY = 'tosca.policies.tacker.Scaling'


class TOSCAToHOT(object):
    """Convert TOSCA template to HOT template."""

    def __init__(self, vnf, heatclient, inst_req_info=None, grant_info=None):
        self.vnf = vnf
        self.heatclient = heatclient
        self.attributes = {}
        self.vnfd_yaml = None
        self.unsupported_props = {}
        self.heat_template_yaml = None
        self.monitoring_dict = None
        self.nested_resources = dict()
        self.fields = None
        self.STACK_FLAVOR_EXTRA = cfg.CONF.openstack_vim.flavor_extra_specs
        self.appmonitoring_dict = None
        self.grant_info = grant_info
        self.inst_req_info = inst_req_info

    @log.log
    def generate_hot(self):

        self._get_vnfd()
        dev_attrs = self._update_fields()
        vnfd_dict = yamlparser.simple_ordered_parse(self.vnfd_yaml)
        LOG.debug('vnfd_dict %s', vnfd_dict)
        self._get_unsupported_resource_props(self.heatclient)

        self._generate_hot_from_tosca(vnfd_dict, dev_attrs,
                                      self.inst_req_info, self.grant_info)
        self.fields['template'] = self.heat_template_yaml
        if not self.vnf['attributes'].get('heat_template'):
            self.vnf['attributes']['heat_template'] = self.fields['template']
        if self.monitoring_dict:
            self.vnf['attributes'][
                'monitoring_policy'] = jsonutils.dump_as_bytes(
                self.monitoring_dict)
        if self.appmonitoring_dict:
            self.vnf['attributes']['app_monitoring_policy'] = \
                jsonutils.dump_as_bytes(self.appmonitoring_dict)

    @log.log
    def _get_vnfd(self):
        self.attributes = self.vnf['vnfd']['attributes'].copy()
        self.vnfd_yaml = self.attributes.pop('vnfd', None)
        if self.vnfd_yaml is None:
            LOG.error("VNFD is not provided, so no vnf is created !!")
            raise exceptions.InvalidInput("VNFD template is None.")
        LOG.debug('vnfd_yaml %s', self.vnfd_yaml)

    @log.log
    def _update_fields(self):
        attributes = self.attributes
        fields = dict((key, attributes.pop(key)) for key
                      in ('stack_name', 'template_url', 'template')
                      if key in attributes)
        for key in ('files', 'parameters'):
            if key in attributes:
                fields[key] = jsonutils.loads(attributes.pop(key))

        # overwrite parameters with given dev_attrs for vnf creation
        dev_attrs = self.vnf['attributes'].copy()
        fields.update(dict((key, dev_attrs.pop(key)) for key
                      in ('stack_name', 'template_url', 'template')
                      if key in dev_attrs))
        for key in ('files', 'parameters'):
            if key in dev_attrs:
                fields.setdefault(key, {}).update(
                    jsonutils.loads(dev_attrs.pop(key)))

        self.attributes = attributes
        self.fields = fields
        return dev_attrs

    @log.log
    def _update_params(self, original, paramvalues, match=False):
        for key, value in (original).items():
            if not isinstance(value, dict) or 'get_input' not in str(value):
                pass
            elif isinstance(value, dict):
                if not match:
                    if key in paramvalues and 'param' in paramvalues[key]:
                        self._update_params(value, paramvalues[key]['param'],
                                            True)
                    elif key in paramvalues:
                        self._update_params(value, paramvalues[key], False)
                    else:
                        LOG.debug('Key missing Value: %s', key)
                        raise cs.InputValuesMissing(key=key)
                elif 'get_input' in value:
                    if value['get_input'] in paramvalues:
                        original[key] = paramvalues[value['get_input']]
                    else:
                        LOG.debug('Key missing Value: %s', key)
                        raise cs.InputValuesMissing(key=key)
                else:
                    self._update_params(value, paramvalues, True)

    @log.log
    def _process_parameterized_input(self, dev_attrs, vnfd_dict):
        param_vattrs_yaml = dev_attrs.pop('param_values', None)
        if param_vattrs_yaml:
            try:
                param_vattrs_dict = yaml.safe_load(param_vattrs_yaml)
                LOG.debug('param_vattrs_yaml', param_vattrs_dict)
            except Exception as e:
                LOG.debug("Not Well Formed: %s", str(e))
                raise vnfm.ParamYAMLNotWellFormed(
                    error_msg_details=str(e))
            else:
                self._update_params(vnfd_dict, param_vattrs_dict)
        else:
            raise cs.ParamYAMLInputMissing()

    @log.log
    def _process_vdu_network_interfaces(self, vdu_id, vdu_dict, properties,
                                        template_dict):

        networks_list = []
        properties['networks'] = networks_list
        for network_param in vdu_dict['network_interfaces'].values():
            port = None
            if 'addresses' in network_param:
                ip_list = network_param.pop('addresses', [])
                if not isinstance(ip_list, list):
                    raise vnfm.IPAddrInvalidInput()
                mgmt_flag = network_param.pop('management', False)
                port, template_dict =\
                    self._handle_port_creation(vdu_id, network_param,
                                               template_dict,
                                               ip_list, mgmt_flag)
            if network_param.pop('management', False):
                port, template_dict = self._handle_port_creation(vdu_id,
                                                                 network_param,
                                                                 template_dict,
                                                                 [], True)
            if port is not None:
                network_param = {
                    'port': {'get_resource': port}
                }
            networks_list.append(dict(network_param))
        return vdu_dict, template_dict

    @log.log
    def _make_port_dict(self):
        port_dict = {'type': 'OS::Neutron::Port'}
        if self.unsupported_props:
            port_dict['properties'] = {
                'value_specs': {
                    'port_security_enabled': False
                }
            }
        else:
            port_dict['properties'] = {
                'port_security_enabled': False
            }
        port_dict['properties'].setdefault('fixed_ips', [])
        return port_dict

    @log.log
    def _make_mgmt_outputs_dict(self, vdu_id, port, template_dict):
        mgmt_ip = 'mgmt_ip-%s' % vdu_id
        outputs_dict = template_dict['outputs']
        outputs_dict[mgmt_ip] = {
            'description': 'management ip address',
            'value': {
                'get_attr': [port, 'fixed_ips', 0, 'ip_address']
            }
        }
        template_dict['outputs'] = outputs_dict
        return template_dict

    @log.log
    def _handle_port_creation(self, vdu_id, network_param,
                              template_dict, ip_list=None,
                              mgmt_flag=False):
        ip_list = ip_list or []
        port = '%s-%s-port' % (vdu_id, network_param['network'])
        port_dict = self._make_port_dict()
        if mgmt_flag:
            template_dict = self._make_mgmt_outputs_dict(vdu_id, port,
                                                         template_dict)
        for ip in ip_list:
            port_dict['properties']['fixed_ips'].append({"ip_address": ip})
        port_dict['properties'].update(network_param)
        template_dict['resources'][port] = port_dict
        return port, template_dict

    @log.log
    def _get_unsupported_resource_props(self, heat_client):
        unsupported_resource_props = {}

        for res, prop_dict in (HEAT_VERSION_INCOMPATIBILITY_MAP).items():
            unsupported_props = {}
            for prop, val in (prop_dict).items():
                if not heat_client.resource_attr_support(res, prop):
                    unsupported_props.update(prop_dict)
            if unsupported_props:
                unsupported_resource_props[res] = unsupported_props
        self.unsupported_props = unsupported_resource_props

    @log.log
    def _generate_hot_from_tosca(self, vnfd_dict, dev_attrs,
                                 inst_req_info=None,
                                 grant_info=None):
        parsed_params = {}
        if 'param_values' in dev_attrs and dev_attrs['param_values'] != "":
            try:
                parsed_params = yaml.safe_load(dev_attrs['param_values'])
            except Exception as e:
                LOG.debug("Params not Well Formed: %s", str(e))
                raise vnfm.ParamYAMLNotWellFormed(error_msg_details=str(e))

        appmonitoring_dict = \
            toscautils.get_vdu_applicationmonitoring(vnfd_dict)

        block_storage_details = toscautils.get_block_storage_details(
            vnfd_dict)
        toscautils.updateimports(vnfd_dict)
        if 'substitution_mappings' in str(vnfd_dict):
            toscautils.check_for_substitution_mappings(
                vnfd_dict,
                parsed_params
            )

        try:
            tosca = tosca_template.ToscaTemplate(parsed_params=parsed_params,
                                                 a_file=False,
                                                 yaml_dict_tpl=vnfd_dict)

        except Exception as e:
            LOG.debug("tosca-parser error: %s", str(e))
            raise vnfm.ToscaParserFailed(error_msg_details=str(e))

        unique_id = uuidutils.generate_uuid()
        metadata = toscautils.get_vdu_metadata(tosca, unique_id=unique_id)
        for policy in tosca.policies:
            if policy.entity_tpl['type'] == constants.POLICY_RESERVATION:
                metadata = toscautils.get_metadata_for_reservation(
                    tosca, metadata)
                break

        alarm_resources = toscautils.pre_process_alarm_resources(
            self.vnf, tosca, metadata, unique_id=unique_id)
        monitoring_dict = toscautils.get_vdu_monitoring(tosca)
        mgmt_ports = toscautils.get_mgmt_ports(tosca)
        res_tpl = toscautils.get_resources_dict(tosca,
                                                self.STACK_FLAVOR_EXTRA)
        toscautils.post_process_template(tosca)
        scaling_policy_names = toscautils.get_scaling_policy(tosca)
        try:
            translator = tosca_translator.TOSCATranslator(tosca, parsed_params)

            heat_template_yaml = translator.translate()
            nested_resource_names = toscautils.get_nested_resources_name(
                heat_template_yaml)
            if nested_resource_names:
                for nested_resource_name in nested_resource_names:
                    sub_heat_tmpl_name = \
                        toscautils.get_sub_heat_tmpl_name(nested_resource_name)
                    sub_heat_template_yaml =\
                        translator.translate_to_yaml_files_dict(
                            sub_heat_tmpl_name)
                    nested_resource_yaml = \
                        sub_heat_template_yaml[nested_resource_name]
                    LOG.debug("nested_resource_yaml: %s", nested_resource_yaml)
                    self.nested_resources[nested_resource_name] = \
                        nested_resource_yaml

        except Exception as e:
            LOG.debug("heat-translator error: %s", str(e))
            raise vnfm.HeatTranslatorFailed(error_msg_details=str(e))

        if self.nested_resources:
            nested_tpl = toscautils.update_nested_scaling_resources(
                self.nested_resources,
                mgmt_ports, metadata, res_tpl, self.unsupported_props,
                grant_info=grant_info, inst_req_info=inst_req_info)
            self.fields['files'] = nested_tpl
            for nested_resource_name in nested_tpl.keys():
                self.vnf['attributes'][nested_resource_name] =\
                    nested_tpl[nested_resource_name]
            mgmt_ports.clear()

        if scaling_policy_names:
            scaling_group_dict = toscautils.get_scaling_group_dict(
                heat_template_yaml, scaling_policy_names)
            self.vnf['attributes']['scaling_group_names'] =\
                jsonutils.dump_as_bytes(scaling_group_dict)

        if self.vnf['attributes'].get('maintenance', None):
            toscautils.add_maintenance_resources(tosca, res_tpl)

        heat_template_yaml = toscautils.post_process_heat_template(
            heat_template_yaml, mgmt_ports, metadata, alarm_resources,
            res_tpl, block_storage_details, self.unsupported_props,
            unique_id=unique_id, inst_req_info=inst_req_info,
            grant_info=grant_info, tosca=tosca)

        try:
            for nested_resource_name in self.nested_resources.keys():
                self.nested_resources[nested_resource_name] = \
                    toscautils.post_process_heat_template_for_scaling(
                    self.nested_resources[nested_resource_name],
                    mgmt_ports, metadata, alarm_resources,
                    res_tpl, block_storage_details, self.unsupported_props,
                    unique_id=unique_id, inst_req_info=inst_req_info,
                    grant_info=grant_info, tosca=tosca)
        except Exception as e:
            LOG.debug("post_process_heat_template_for_scaling "
                      "error: %s", str(e))
            raise

        self.heat_template_yaml = heat_template_yaml
        self.monitoring_dict = monitoring_dict
        self.metadata = metadata
        self.appmonitoring_dict = appmonitoring_dict

    @log.log
    def represent_odict(self, dump, tag, mapping, flow_style=None):
        value = []
        node = yaml.MappingNode(tag, value, flow_style=flow_style)
        if dump.alias_key is not None:
            dump.represented_objects[dump.alias_key] = node
        best_style = True
        if hasattr(mapping, 'items'):
            mapping = mapping.items()
        for item_key, item_value in mapping:
            node_key = dump.represent_data(item_key)
            node_value = dump.represent_data(item_value)
            if not (isinstance(node_key, yaml.ScalarNode)
                    and not node_key.style):
                best_style = False
            if not (isinstance(node_value, yaml.ScalarNode)
                    and not node_value.style):
                best_style = False
            value.append((node_key, node_value))
        if flow_style is None:
            if dump.default_flow_style is not None:
                node.flow_style = dump.default_flow_style
            else:
                node.flow_style = best_style
        return node



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from tacker.common.rpc import BackingOffClient
from tacker.conductor.conductorrpc import vnf_pkgm_rpc
from tacker.objects import vnf_package
from tacker.tests import base
from tacker.tests.unit.conductor import fakes


class VnfPackageRPCTestCase(base.BaseTestCase):

    def setUp(self):
        super(VnfPackageRPCTestCase, self).setUp()
        self.context = self.fake_admin_context()
        self.rpc_api = vnf_pkgm_rpc.VNFPackageRPCAPI()
        self.cctxt_mock = mock.MagicMock()

    def test_upload_vnf_package_content(self):

        @mock.patch.object(BackingOffClient, 'prepare')
        def _test(prepare_mock):
            prepare_mock.return_value = self.cctxt_mock
            vnf_package_obj = vnf_package.VnfPackage(
                self.context, **fakes.VNF_UPLOAD_VNF_PACKAGE_CONTENT)
            self.rpc_api.upload_vnf_package_content(self.context,
                                            vnf_package_obj, cast=True)
            prepare_mock.assert_called()
            self.cctxt_mock.cast.assert_called_once_with(
                self.context, 'upload_vnf_package_content',
                vnf_package=vnf_package_obj)
        _test()

    def test_upload_vnf_package_from_uri(self):
        fake_addressInformation = "http://test_csar.zip"

        @mock.patch.object(BackingOffClient, 'prepare')
        def _test(prepare_mock):
            prepare_mock.return_value = self.cctxt_mock
            vnf_package_obj = vnf_package.VnfPackage(self.context,
                                                     **fakes.VNF_DATA)
            self.rpc_api.upload_vnf_package_from_uri(self.context,
                                                     vnf_package_obj,
                                                     fake_addressInformation,
                                                     cast=True)
            prepare_mock.assert_called()
            self.cctxt_mock.cast.assert_called_once_with(
                self.context, 'upload_vnf_package_from_uri',
                vnf_package=vnf_package_obj,
                address_information=fake_addressInformation,
                password=None, user_name=None)
        _test()

    def test_delete_vnf_package(self):

        @mock.patch.object(BackingOffClient, 'prepare')
        def _test(prepare_mock):
            prepare_mock.return_value = self.cctxt_mock
            vnf_package_obj = vnf_package.VnfPackage(self.context,
                                                     **fakes.VNF_DATA)
            self.rpc_api.delete_vnf_package(self.context,
                                            vnf_package_obj, cast=True)
            prepare_mock.assert_called()
            self.cctxt_mock.cast.assert_called_once_with(
                self.context, 'delete_vnf_package',
                vnf_package=vnf_package_obj)
        _test()

    def test_get_vnf_package_vnfd(self):

        @mock.patch.object(BackingOffClient, 'prepare')
        def _test(prepare_mock):
            prepare_mock.return_value = self.cctxt_mock
            vnf_package_obj = vnf_package.VnfPackage(self.context,
                                                     **fakes.VNF_DATA)
            self.rpc_api.get_vnf_package_vnfd(self.context,
                                            vnf_package_obj, cast=False)
            prepare_mock.assert_called()
            self.cctxt_mock.call.assert_called_once_with(
                self.context, 'get_vnf_package_vnfd',
                vnf_package=vnf_package_obj)
        _test()



# Copyright (C) 2020 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import fixtures
import os
import shutil

from oslo_config import cfg
from oslo_utils import uuidutils

from tacker.common import exceptions
from tacker import context
from tacker.manager import TackerManager
from tacker import objects
from tacker.objects import fields
from tacker.tests.unit.db import base as db_base
from tacker.tests.unit.nfvo.test_nfvo_plugin import FakeVNFMPlugin
from tacker.tests.unit.vnflcm import fakes
from tacker.tests import utils as test_utils
from tacker.tests import uuidsentinel
from tacker.vnflcm import vnflcm_driver
from tacker.vnflcm.vnflcm_driver import VnfLcmDriver
from unittest import mock


vnf_dict = {
    'id': uuidutils.generate_uuid(),
    'mgmt_ip_address': '{"VDU1": "a.b.c.d"}',
    'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
    'instance_id': 'a737497c-761c-11e5-89c3-9cb6541d805d',
    'vnfd': {
        'attributes': {
            'heat_template': {
                'resources': {
                    'VDU1': {
                        'properties': {
                            'networks': [{'port': {'get_resource': 'CP1'}}]}
                    }
                }
            }
        }
    }
}


class InfraDriverException(Exception):
    pass


class FakeDriverManager(mock.Mock):
    def __init__(self, fail_method_name=None, vnf_resource_count=1):
        super(FakeDriverManager, self).__init__()
        self.fail_method_name = fail_method_name
        self.vnf_resource_count = vnf_resource_count

    def invoke(self, *args, **kwargs):
        if 'pre_instantiation_vnf' in args:
            vnf_resource_list = [fakes.return_vnf_resource() for index in
                range(self.vnf_resource_count)]
            return {'node_name': vnf_resource_list}
        if 'instantiate_vnf' in args:
            if self.fail_method_name and \
                    self.fail_method_name == 'instantiate_vnf':
                raise InfraDriverException("instantiate_vnf failed")

            instance_id = uuidsentinel.instance_id
            vnfd_dict = kwargs.get('vnfd_dict')
            vnfd_dict['instance_id'] = instance_id
            return instance_id
        if 'create_wait' in args:
            if self.fail_method_name and \
                    self.fail_method_name == 'create_wait':
                raise InfraDriverException("create_wait failed")
        elif 'post_vnf_instantiation' in args:
            pass
        if 'mgmt-drivers-custom' in args:
            if self.fail_method_name and \
                    self.fail_method_name == 'mgmt-drivers-custom':
                raise InfraDriverException("mgmt-drivers-custom failed")


class FakeVimClient(mock.Mock):
    pass


class FakeTackerManager(mock.MagicMock):
    pass


class MgmtVnfLcmDriverTest(db_base.SqlTestCase):

    def setUp(self):
        super(MgmtVnfLcmDriverTest, self).setUp()
        self.addCleanup(mock.patch.stopall)
        self.context = context.get_admin_context()
        self._mock_vim_client()
        self._stub_get_vim()
        self.temp_dir = self.useFixture(fixtures.TempDir()).path

    def _mock_vnf_manager(self, fail_method_name=None, vnf_resource_count=1):
        self._vnf_manager = mock.Mock(wraps=FakeDriverManager(
            fail_method_name=fail_method_name,
            vnf_resource_count=vnf_resource_count))
        self._vnf_manager.__contains__ = mock.Mock(
            return_value=True)
        fake_vnf_manager = mock.Mock()
        fake_vnf_manager.return_value = self._vnf_manager
        self._mock(
            'tacker.common.driver_manager.DriverManager', fake_vnf_manager)

    def _mock_vim_client(self):
        self.vim_client = mock.Mock(wraps=FakeVimClient())
        fake_vim_client = mock.Mock()
        fake_vim_client.return_value = self.vim_client
        self._mock(
            'tacker.vnfm.vim_client.VimClient', fake_vim_client)

    def _stub_get_vim(self):
        vim_obj = {'vim_id': '6261579e-d6f3-49ad-8bc3-a9cb974778ff',
                   'vim_name': 'fake_vim',
                   'vim_auth':
                       {'auth_url': 'http://localhost/identity',
                        'password': 'test_pw', 'username': 'test_user',
                        'project_name': 'test_project'},
                   'vim_type': 'openstack',
                   'tenant': uuidsentinel.tenant_id}
        self.vim_client.get_vim.return_value = vim_obj

    @mock.patch('tacker.vnflcm.utils.get_default_scale_status')
    @mock.patch('tacker.vnflcm.utils._make_final_vnf_dict')
    @mock.patch.object(VnfLcmDriver, '_init_mgmt_driver_hash')
    @mock.patch.object(TackerManager, 'get_service_plugins',
        return_value={'VNFM': FakeVNFMPlugin()})
    @mock.patch.object(objects.VnfResource, 'create')
    @mock.patch.object(objects.VnfPackageVnfd, 'get_by_id')
    @mock.patch.object(objects.VnfInstance, "save")
    def test_instantiate_vnf(self, mock_vnf_instance_save,
                             mock_vnf_package_vnfd, mock_create,
                             mock_get_service_plugins, mock_init_hash,
                             mock_final_vnf_dict, mock_default_status):
        mock_init_hash.return_value = {
            "vnflcm_noop": "ffea638bfdbde3fb01f191bbe75b031859"
                           "b18d663b127100eb72b19eecd7ed51"
        }
        vnf_package_vnfd = fakes.return_vnf_package_vnfd()
        vnf_package_id = vnf_package_vnfd.package_uuid
        mock_vnf_package_vnfd.return_value = vnf_package_vnfd
        instantiate_vnf_req_dict = fakes.get_dummy_instantiate_vnf_request()
        instantiate_vnf_req_obj = \
            objects.InstantiateVnfRequest.obj_from_primitive(
                instantiate_vnf_req_dict, self.context)
        vnf_instance_obj = fakes.return_vnf_instance()
        mock_default_status.return_value = None

        fake_csar = os.path.join(self.temp_dir, vnf_package_id)
        cfg.CONF.set_override('vnf_package_csar_path', self.temp_dir,
                              group='vnf_package')
        test_utils.copy_csar_files(fake_csar, "refactor_mgmt_driver1")
        self._mock_vnf_manager()
        driver = vnflcm_driver.VnfLcmDriver()
        vnf_dict = {
            "vnfd": {"attributes": {}}, "attributes": {},
            "before_error_point": fields.ErrorPoint.VNF_CONFIG_START}
        driver.instantiate_vnf(self.context, vnf_instance_obj, vnf_dict,
                               instantiate_vnf_req_obj)

        self.assertEqual(1, mock_vnf_instance_save.call_count)
        self.assertEqual(6, self._vnf_manager.invoke.call_count)
        shutil.rmtree(fake_csar)

    @mock.patch('tacker.vnflcm.utils._make_final_vnf_dict')
    @mock.patch.object(VnfLcmDriver, '_init_mgmt_driver_hash')
    @mock.patch.object(TackerManager, 'get_service_plugins',
                       return_value={'VNFM': FakeVNFMPlugin()})
    @mock.patch.object(objects.VnfResource, 'create')
    @mock.patch.object(objects.VnfPackageVnfd, 'get_by_id')
    @mock.patch.object(objects.VnfInstance, "save")
    def test_instantiate_vnf_py_name_false(
            self, mock_vnf_instance_save, mock_vnf_package_vnfd,
            mock_create, mock_get_service_plugins, mock_init_hash,
            mock_final_vnf_dict):
        mock_init_hash.return_value = {
            "vnflcm_noop": "ffea638bfdbde3fb01f191bbe75b031859"
                           "b18d663b127100eb72b19eecd7ed51"
        }
        vnf_package_vnfd = fakes.return_vnf_package_vnfd()
        vnf_package_id = vnf_package_vnfd.package_uuid
        mock_vnf_package_vnfd.return_value = vnf_package_vnfd
        instantiate_vnf_req_dict = fakes.get_dummy_instantiate_vnf_request()
        instantiate_vnf_req_obj = \
            objects.InstantiateVnfRequest.obj_from_primitive(
                instantiate_vnf_req_dict, self.context)
        vnf_instance_obj = fakes.return_vnf_instance()

        fake_csar = os.path.join(self.temp_dir, vnf_package_id)
        cfg.CONF.set_override('vnf_package_csar_path', self.temp_dir,
                              group='vnf_package')
        test_utils.copy_csar_files(fake_csar, "refactor_mgmt_driver2")
        self._mock_vnf_manager()
        driver = vnflcm_driver.VnfLcmDriver()
        vnf_dict = {
            "vnfd": {"attributes": {}}, "attributes": {},
            "before_error_point": fields.ErrorPoint.VNF_CONFIG_START}
        self.assertRaises(exceptions.MgmtDriverInconsistent,
                          driver.instantiate_vnf, self.context,
                          vnf_instance_obj, vnf_dict,
                          instantiate_vnf_req_obj)
        shutil.rmtree(fake_csar)

    @mock.patch('tacker.vnflcm.utils._make_final_vnf_dict')
    @mock.patch.object(VnfLcmDriver, '_init_mgmt_driver_hash')
    @mock.patch.object(TackerManager, 'get_service_plugins',
                       return_value={'VNFM': FakeVNFMPlugin()})
    @mock.patch.object(objects.VnfResource, 'create')
    @mock.patch.object(objects.VnfPackageVnfd, 'get_by_id')
    @mock.patch.object(objects.VnfInstance, "save")
    def test_instantiate_vnf_py_hash_false(
            self, mock_vnf_instance_save, mock_vnf_package_vnfd,
            mock_create, mock_get_service_plugins, mock_init_hash,
            mock_final_vnf_dict):
        mock_init_hash.return_value = {
            "vnflcm_noop": "ffea638bfdbde3fb01f191bbe75b031859"
                           "b18d663b127100eb72b19eecd7ed51"
        }
        vnf_package_vnfd = fakes.return_vnf_package_vnfd()
        vnf_package_id = vnf_package_vnfd.package_uuid
        mock_vnf_package_vnfd.return_value = vnf_package_vnfd
        instantiate_vnf_req_dict = fakes.get_dummy_instantiate_vnf_request()
        instantiate_vnf_req_obj = \
            objects.InstantiateVnfRequest.obj_from_primitive(
                instantiate_vnf_req_dict, self.context)
        vnf_instance_obj = fakes.return_vnf_instance()

        fake_csar = os.path.join(self.temp_dir, vnf_package_id)
        cfg.CONF.set_override('vnf_package_csar_path', self.temp_dir,
                              group='vnf_package')
        test_utils.copy_csar_files(fake_csar, "refactor_mgmt_driver3")
        self._mock_vnf_manager()
        driver = vnflcm_driver.VnfLcmDriver()
        vnf_dict = {
            "vnfd": {"attributes": {}}, "attributes": {},
            "before_error_point": fields.ErrorPoint.VNF_CONFIG_START}
        self.assertRaises(exceptions.MgmtDriverHashMatchFailure,
                          driver.instantiate_vnf, self.context,
                          vnf_instance_obj, vnf_dict,
                          instantiate_vnf_req_obj)
        shutil.rmtree(fake_csar)
